/*
    This file is part of the UNF library / toolset
    Copyright (C) 2000 Steffen Kaiser

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
/* $RCSfile: unf.h $
   $Locker: ska $	$Name:  $	$State: Exp $

	Global UNF declarations

*/

#ifndef SK__UNF_H_
#define SK__UNF_H_

#include <stdio.h>
#include <stdarg.h>

	/* UNF uses the same structures as the kernel does */
#include "nls.loc"

#define UNF_RDX_COMBINED (-1)	/* Combined radix;	Output uses decimal */
#define UNF_RDX_CSTYLE (-2)		/* C-style radix;	Output uses decimal */

#define UNF_DEFAULT (-1)

	/* I used to assume that BYTE & (D)WORD are _unsigned_ rather than
		signed, but the FreeDOS kernel does not agree. -- 2000/05/01 ska*/
typedef unsigned long UDWORD;


	/* one byte alignment */
#include <algnbyte.h>

	/* This structure is a dummy structure to poll information from
		DOS-65-XX */
struct UNF_IN_BUFFER {
	UBYTE subFct;
	union {
		UBYTE FAR *pointer;
		struct {
			UWORD size;
			UWORD code;
			UWORD cp;
			UBYTE buf[36];
		} subInfo;
	} info;
};


struct UNF_DATA_ENTRY {
#if 1
    SHORT fct;          /* DOS-65-XX subfct number */
    UWORD size;         /* size of the _complete_ data block */
#else
	int fct;			/* DOS-65-XX subfct number */
	unsigned size;		/* size of the _complete_ data block */
#endif
	UBYTE data[1];		/* dynamically extended data block */
};


struct UNF_DATA {		/* a complete UNF data block */
	UWORD country, codepage;		/* IDs of loaded NLS data */
#if 1
    SHORT numData;      /* number of entries within array */
#else
	int numData;		/* number of entries within array */
#endif
	struct UNF_DATA_ENTRY *data[1]; /* dynamically extended.
							The pointers are dynamically allocated memory
							blocks. There is _no_ specific order of the
							entries. There is one entry per subfct
							at maximum. */
};

struct UNF_IFile {			/* input file context */
	struct UNF_IFile *prv;
	unsigned long linenr;
	FILE *fp;
	char fname[1];
};

/* Check if a character is viewable.
	Usually the standard <ctype.h> functions supports "C" locale only
	within DOS, but often excluding characters above code 127 completely.
*/
int unf_isviewable(int ch);
	/* Allocate memory and terminate on failure */
void *unf_alloc(unsigned size);
char *unf_strdup(char *);
void *unf_realloc(void *p, unsigned size);
void unf_free(void *p);
char *unf_stpcpy(char *, const char *); /* 20140822 sava */

/* error reporting callback function */
typedef enum {
	 UNF_ENone = 0
	,UNF_EOutOfMemory		/* malloc() (and assoc.) failed */
	,UNF_ENumArg			/* invalid numerical argument */
	,UNF_EOpenFile
	,UNF_ENumExceeded		/* both cntry ID & cp exceed 999 */
	,UNF_ELineTooLong
	,UNF_EInvalidOpt		/* syntactically invalid option */
	,UNF_EUnknownOpt		/* semantically unknwon option */
	,UNF_ERadix				/* radix exceeds range */
	,UNF_EOffMismatch		/* $is_offset mismatch */
	,UNF_EMisplacedData		/* data line not allowed at this position */
	,UNF_EDataline			/* syntax error in data line */
	,UNF_EMissingBasedata	/* missing country ID and/or codepage ID */
	,UNF_EInvalidTableName
	,UNF_ESyntax			/* general syntax error in line */
	,UNF_EInvalidBasedata	/* country ID and/or codepage ID invalid */
	,UNF_ENoTables			/* no single table specified */
	,UNF_EMissingTable		/* Missing required table */
	,UNF_ESizeTable			/* Size of standard table is wrong */
	,UNF_EValue				/* Value of table invalid */
	,UNF_EStringValue		/* String value of table not '\0' terminated */
	,UNF_ERangeValue		/* Range values out of range or begin > end */
	,UNF_EFTermChars		/* Filename terminator characters invalid */
	,UNF_EDuplicateTable	/* Duplicated table */
	,UNF_EInvalidTable		/* Invalid table ID (0x20..0x22, 0xA0..0xA2) */
} UNF_Error_ID;

typedef void UNF_Error_Fct(struct UNF_IContext * const, UNF_Error_ID
	, char const * const fmt, ...);
struct UNF_IContext {	/* input context */
	int radix;				/* current input radix */
	UNF_Error_ID error_id;	/* ID of first error occured */
	UNF_Error_Fct *error_fct;	/* function to report errors to */
	struct UNF_IFile *fin;
};

void unf_error(struct UNF_IContext * const, UNF_Error_ID
	, char const * const fmt, ...);
void unf_dummy_error(struct UNF_IContext * const, UNF_Error_ID
	, char const * const fmt, ...);
	/* This function is used to terminate the program if the output functions
		detect a fatal error */
void unf__eerror(char const * const fmt, ...);


/**
 **	Output functions
 **/
void unf_oc(int ch);			/* output character */
void unf_os(char *s);			/* output string */
void unf_of(char *fmt, ...);	/* output  la printf() */
void unf_onl(void);				/* output end-of-line */
void unf_ocmt(char *s);			/* output comment & end-of-line */
void unf_oonam(char *opt);		/* output option name */
void unf_oo(char *opt);			/* output option without argument */
#define unf_onum(v) unf_ofnum(0, (v))	/* output a number in output radix */
							/* as onum(), but prefix spaces to output
								length characters minimum */
void unf_ofnum(int length, unsigned long value);
void unf_oou_(char *opt, UWORD val);	/* output option with UWORD argument */
void unf_oou(char *opt, UWORD val);	/* output option with UWORD argument */
							/* output option with UWORD argument & comment */
void unf_oouc(char *opt, unsigned val, char *comment);
void unf_olabelc(int table, char *comment);	/* output table label with cmt */
void unf_oend(void);			/* close table label */
void unf_otopen(void);			/* output table open sequence */
void unf_otbf(int value);		/* output table byte formatted */
void unf_otb(int value);		/* output table byte (unformatted) */
void unf_otw(UWORD value);		/* output table word (unformatted) */
void unf_otiw(UWORD value);		/* output table item word */
void unf_otd(UDWORD value);		/* output table dword (unformatted) */
void unf_otid(UDWORD value);		/* output table item dword */
void unf_otp(void FAR * value);	/* unformatted dword */
#define unf_otc(comment)	unf_ocmt(comment)	/* output table comment */
void unf_otlb(int value, char *comment);	/* output table line one byte & cmt */
void unf_otlw(UWORD value, char *comment);	/* output table line one word & cmt */
void unf_otld(UDWORD value, char *comment);	/* output table line one dword & cmt */
							/* output table from array with comment */
#define unf_otac(array,cmt)	unf_ottc(array, sizeof(array), cmt)
							/* output table without comment */
#define unf_ott(buf,len)	unf_ottc((buf), (len), NULL)
							/* output table with comment */
void unf_ottc(unsigned char *buf, size_t len, char *comment);
							/* output table without comment but with
								trailing "is_offset" option */
void unf_ot(unsigned char *buf, size_t len);

							/* output a table #1 / Extended Country Info */
void unf_oExtInf(struct nlsExtCntryInfo *nfo);

							/* output a table #0x23 / Yes/No characters */
void unf_oYesNo(char *ch);

							/* output a character buffer table of function
								'fct' with name 'name' */
void unf_oCharTab(int fct, char *name, UWORD *tab);

							/* output a table #5 / filename term chars */
void unf_oFTerm(struct nlsFnamTerm *nls);

void unf_oDBCS(UWORD *size);		/* output a table #7 / DBCS */
							/* Attention: Supports "no DBCS" only! */

void unf_oradix(int radix);		/* Change output radix */


							/* Create UNF file in given path of given
								country ID & codepage */
int unf_create(char const *path, UWORD cntry, UWORD cp);
#if 1
int unf_creat2(char const *pathname, UWORD cntry, UWORD cp);
#endif
void unf_close(void);
		/* Dump all the information within the UNF_DATA structure into
			a new UNF file using automatically calling
			unf_create(), all necessary special output functions
			for the tables 1..7,& 0x23 and call unf_close(). */
void unf_dump0(int ispathname, struct UNF_DATA const * const d, char const *path);
void unf_dump(struct UNF_DATA const * const d, char const *path);



							/* Scan UNF file in given path of given
								country ID & codepage;
								If "path" is already a file rather than
								a path, no changes are made to it.
								Return is a pointer to a allocated structure;
								all its entries are dyn allocated, too.
								If fct == NULL, unf_load() will return NULL
								on error, but performs no further actions.
								*/
struct UNF_DATA *unf_load(char *path, UWORD cntry, UWORD cp, UNF_Error_Fct *fct);
	/* Assumes that "p" points to a dynamically allocated UNF_DATA structure
		and deallocates all resources including "p" itself */
void unf_freeData(struct UNF_DATA *p);

int unf_sanityCheck(struct UNF_DATA *d, UNF_Error_Fct *fct);
int unf_kernelCheck(struct UNF_DATA *d, UNF_Error_Fct *fct);
	/* Checks of a loaded UNF data stucture.
		sanityCheck() provides minimum checking, e.g. missing standard
		tables.
		kernelCheck() tests if the NLS data can be handled by the
		kernel-internal code

		Return: The number of flaws
	*/

int *unf_tab(struct UNF_DATA *d);
	/* Returns an array of 256 items mapping the table_id to the
		index of the table within the d->data[] array.
		Absent tables are indicated by (int)-1.
		If d->data[] contains the same table twice, the last one
		superceeds all previous ones. */
int unf_get(int subfct, struct UNF_IN_BUFFER *iBuf);
	/* Retreives the data from DOS-65-XX (XX == subfct) and stores
		it into *iBuf. The buffer is filled with exactly the same
		information as returned by DOS, that means that it is
		a far pointer except when subfct == 1.
		Return:
			0: On success; *iBuf updated
			else: on error: *iBuf cleared, system error code
	*/

#endif
