/*
    This file is part of the UNF library / toolset
    Copyright (C) 2000 Steffen Kaiser

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
/* $RCSfile: UNF_CRT.C $
   $Locker: ska $	$Name:  $	$State: Exp $

	Create UNF file.

	If an output file had been opened already, it is closed.

	Input: Country ID & Codepage
		'path' maybe NULL
	Output: 0: on success
			1: fopen() failed
			2: malloc() failed

*/

#include "unf.loc"

#ifndef _MICROC_
#include <stdlib.h>
#include <string.h>
#endif

#include "unf_post.loc"

#ifdef RCS_Version
static char const rcsid[] = 
	"$Id: UNF_CRT.C 1.5 2000/04/29 04:44:46 ska Exp ska $";
#endif

int unf_create(char const *path, UWORD cntry, UWORD cp)
{	char *buf, *p;

#if 1
    if(unf_out) {
        int rc = fclose(unf_out);
        unf_out = NULL;
        if (rc != 0)
            return 1;
    }
#else
	if(unf_out)
		fclose(unf_out);
#endif

	if(!path)
		path = "";
	buf = unf_alloc(strlen(path) + 1 + 2*(sizeof(cntry) * 8 + 1) + 1 + 4);
	if(buf != NULL) {
		if(*path) {
			p = unf_stpcpy(buf, path);
			if(p[-1] != '/' && p[-1] != '\\' && p[-1] != ':')
				*p++ = '\\';
		} else p = buf;
		if(cntry > 999 && cp > 999)
			unf__eerror("Both country ID and codepage exceed 999");
		sprintf(p, "%03u-%03u.unf", cntry, cp);
		unf_out = fopen(buf, "wt");
		if(unf_out) {
			of("## Universal NLS data Format file: %s\n$country %u\n$codepage %u\n\n# Global options\n", p, cntry, cp);
			orx = 10;		/* Default radix */
		}
		free(buf);
		return unf_out == 0;
	}
	return 2;
}

#if 1
int unf_creat2(const char *pathname, UWORD cntry, UWORD cp)
{
    if(unf_out && unf_out != stdout) {
        int rc = fclose(unf_out);
        unf_out = NULL;
        if (rc != 0)
            return 1;
    }
    if (pathname && (strcmp(pathname, "--") == 0 || strcmp(pathname, "stdout") == 0)) {
        unf_out = stdout;
    }
    else
        unf_out = fopen(pathname, "wt");
    if (!unf_out)
         unf__eerror("Can't open file %s.\n", pathname);
    of("## Universal NLS data Format file: %s\n$country %u\n$codepage %u\n\n# Global options\n", pathname, cntry, cp);
    
    return 0;
}
#endif
