/*
    This file is part of the UNF library / toolset
    Copyright (C) 2000 Steffen Kaiser

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
/* $RCSfile: UNF_KCHK.C $
   $Locker: ska $	$Name:  $	$State: Exp $

	Checking of given NLS data if it is compatible with the kernel code

	Checks for:
	+ sanityCheck()
	+ standard table size mismatches (of standard tables only)

	Return:
		The number of flaws

*/

#include "unf.loc"

#ifdef _MICROC_
#endif
#include <string.h>

#include "unf_post.loc"

#ifdef RCS_Version
static char const rcsid[] = 
	"$Id: UNF_KCHK.C 1.2 2000/06/17 22:32:11 ska Exp ska $";
#endif

#define unf_error ++flaws; (*error_fct)( NULL, 


int unf_kernelCheck(struct UNF_DATA *d, UNF_Error_Fct *efct)
{	int flaws;					/* number of problems */
	UNF_Error_Fct *error_fct;	/* function to report errors to */
	int i, fct;

	error_fct = efct? efct: (UNF_Error_Fct*)unf_dummy_error;

	if((flaws = unf_sanityCheck(d, efct)) != 0)
		return flaws;


	i = d->numData;
	while(i--) if(d->data[i]) switch(fct = d->data[i]->fct) {
	case 0x23:	/* Check table 0x23 Yes/No */
	case 1: /* Checking table 1 Ext Cntry Info */
		/* sanity check did everything */
		/* add new fields here */
		break;

	case 2:	/* Check table 2 upcase */
	case 4:	/* Check table filename 4 upcase */
		/* must contain exactly 128 characters */
		if(d->data[i]->size != 128 + 2
		 || *(UWORD*)(d->data[i]->data) != 128) {
			unf_error UNF_ESizeTable, "Size of table %u is invalid", fct);
		}
		break;

	case 5:	/* Check table 5 filename terminator */
		/* sanityCheck() verified first/last pairs. Now make sure the
			separators[] contain some standard characters */
	{	struct nlsFnamTerm *fn;

		fn = (struct nlsFnamTerm*)(d->data[i]->data);
		if(!fn->numSep
		 || !memchr(fn->separators, ':', fn->numSep)
		 || !memchr(fn->separators, '\\', fn->numSep)
		 || !memchr(fn->separators, '/', fn->numSep)
		 || !memchr(fn->separators, '"', fn->numSep)
		 || !memchr(fn->separators, '|', fn->numSep)
		 || !memchr(fn->separators, '<', fn->numSep)
		 || !memchr(fn->separators, '>', fn->numSep)) {
		 	unf_error UNF_EFTermChars, "Filename terminator characters must exclude at least ':\\/\"|<>'");
		}
	}
		break;
		
	case 3:	/* Check table 3 lowercase */
	case 6:	/* Check table 6 collating sequenze */
		/* must contain exactly 256 characters */
		if(d->data[i]->size != 256 + 2
		 || *(UWORD*)(d->data[i]->data) != 256) {
			unf_error UNF_ESizeTable, "Size of table %u is invalid", fct);
		}
		break;

	case 0x00:								/* modify data */
	case 0x20: case 0x21: case 0x22:		/* normal character upcase */
	case 0xA0: case 0xA1: case 0xA2:		/* filename character upcase */
		/* These tables are invalid, because a function generates their
			result(s). */
		unf_error UNF_EInvalidTable, "Invalid table %u", fct);
		break;

	/* no other table is checked at all */
	}


#if 0
errRet:
#endif
	return flaws;
}
