/*
    This file is part of the UNF library / toolset
    Copyright (C) 2000 Steffen Kaiser

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
/* $RCSfile: UNF_LOAD.C $
   $Locker: ska $	$Name:  $	$State: Exp $

	Scan an UNF file into memory; terminate the process on failure.

*/

#include "unf.loc"

#ifndef _MICROC_
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#endif

#include "unf_post.loc"

#ifdef RCS_Version
static char const rcsid[] = 
	"$Id: UNF_LOAD.C 1.2 2000/04/29 04:44:47 ska Exp ska $";
#endif

#define UNF_BUFSIZ 512
#define MAX_OPT_SIZE 32

/* Bitfield of flags */
#define GLOBAL 1
#define LOCAL 2
#define READONLY 4

#define IFile UNF_IFile
#define IContext UNF_IContext

#define itemsof(array)	(sizeof(array) / sizeof(array[0]))

static struct IContext ictxt;

#define irx 	ictxt.radix
#define f 		ictxt.fin->fp
#define	fnam	ictxt.fin->fname
#define line	ictxt.fin->linenr
#define iferr	if(ictxt.error_id != UNF_ENone)
#define ifnerr	if(ictxt.error_id == UNF_ENone)
#define errid	ictxt.error_id
#define unf_error	(*ictxt.error_fct)(&ictxt,

typedef enum { OPT_Unknown
	, OPT_Radix, OPT_Country, OPT_Codepage
	, OPT_Include, OPT_ChainTo
	, OPT_IsOffset, OPT_MakeSize, OPT_StopSize
	, OPT_Goto, OPT_GoBack, OPT_GoForward, OPT_GoEven, OPT_GoOdd
} OptionID;

static struct Options {
	char *name;
	int flags;
	OptionID id;
} opts[] = {
	 { "radix",			GLOBAL | LOCAL ,	OPT_Radix }
	,{ "country",		GLOBAL ,			OPT_Country }
	,{ "codepage",		GLOBAL ,			OPT_Codepage }
	,{ "include",		GLOBAL | LOCAL ,	OPT_Include }
	,{ "chain_to",		GLOBAL | LOCAL ,	OPT_ChainTo }
	,{ "is_offset",		LOCAL ,				OPT_IsOffset }
	,{ "make_size",		LOCAL ,				OPT_MakeSize }
	,{ "stop_size",		LOCAL ,				OPT_StopSize }
	,{ "goto",			LOCAL ,				OPT_Goto }
	,{ "go_back",		LOCAL ,				OPT_GoBack }
	,{ "go_forward",	LOCAL ,				OPT_GoForward }
	,{ "go_even",		LOCAL ,				OPT_GoEven }
	,{ "go_odd",		LOCAL ,				OPT_GoOdd }
};

static char *skipws(char *p)
{	while(isspace(*p))
		++p;
	return p;
}

static char *inum(char *p, unsigned long *num)
{	
	int ch, rx;
	char *q;

	if(!isdigit(*(p = skipws(p))))
		return p;

	q = NULL;
	switch(rx = irx) {
	case UNF_RDX_CSTYLE:
		if(*p == '0') {
			if(toupper(p[1]) == 'X') {
	hex16:
				rx = 16;
				p += 2;
				if(!isalnum(*p))
					return p - 2;
			} else rx = 8;
		} else rx = 10;
		break;
	case UNF_RDX_COMBINED:
		if(*p == '0' && toupper(p[1]) == 'X')
			goto hex16;
		else {
			q = p;
			while(isalnum(*++q));
			switch(*--q) {
			case 'O': case 'o': case 'q': case 'Q':
				rx = 8; break;
			case 'h': case 'H':
				rx = 16; break;
			case 'd': case 'D':
				rx = 10; break;
			case 'b': case 'B':
				rx = 2; break;
			default:
				rx = 10;
				q = NULL;
			}
		}
	}

	*num = 0;
	do {
		if((ch = toupper(*p)) < '0')
			break;
		if((ch -= '0') > 9) {
			if(ch < 'A' - '0' || ch > 'Z' - '0')
				return p;
			ch += '0' - 'A' + 10;
		}

		if(ch >= rx)
			break;
		*num = *num * rx + ch;
	} while(*++p && (!q || p < q));

	if(q == p) ++p;

	return p;
}

#define getWord(var)	numWord(p, (int*)&(var))
static void numWord(char *p, int *num)
{	unsigned long n;

	if(*skipws(inum(p, &n)))
		unf_error UNF_ENumArg, "Invalid numerical argument");

	*num = (int)n;
}

#define addValue(size,num,value)	\
	addToBuf(&e, numSizes, sizes, &offset, size, num, value)
static void addToBuf(struct UNF_DATA_ENTRY **Xe, int numSizes, unsigned *sizes
	, unsigned *offset, int size, int num, unsigned long value)
{	int diff, length;
	struct UNF_DATA_ENTRY *e;
	int i;

	iferr return;

	if((diff = length = num * size) == 0)
		return;			/* nothing to do */

	e = *Xe;
	diff -= e->size - *offset;
	if(diff > 0) {
		*Xe = 
		e = unf_realloc(e, sizeof(struct UNF_DATA_ENTRY) + (e->size += diff));
		for(i = 0; i < numSizes; ++i)
			*(UWORD*)&(e->data[sizes[i]]) += diff;
	}
	if(size == 1)
		memset(&e->data[*offset], (UBYTE)value, num);
	else {
		unsigned long *p;

		if(size == 2) {				/* Advance to DWord */
			value |= value << 16;
			i = num & 1;
			num >>= 1;
		}
		else i = 0;

		p = (unsigned long*)&(e->data[*offset]);
		while(num--)
			*p++ = value;
		if(i)						/* Odd word count */
			*(UWORD*)p = (UWORD)value;
	}
	*offset += length;
}

#if 1
static void expandZeroFillBuf(struct UNF_DATA_ENTRY **Xe, int numSizes, unsigned *sizes, unsigned offset)
{
    unsigned off_datalast;
    if (offset > (*Xe)->size) {
        off_datalast = (*Xe)->size;
        addToBuf(Xe, numSizes, sizes, &off_datalast, 1, offset - (*Xe)->size, 0);
    }
}
#define expandBuf(offset)	\
	expandZeroFillBuf(&e, numSizes, sizes, offset)
#endif

static int popFile(void)
{	struct IFile *p;

	if((p = ictxt.fin) == NULL)
		return 0;

	if(f) fclose(f);
	ictxt.fin = p->prv;
	free(p);
	return ictxt.fin != NULL;
}

static int probeFile(char *fname)
{	struct IFile *p;
	char *logname;

	iferr return 2;

	logname = strcmp(fname, "-")? fname: "<<stdin>>";

	p = unf_alloc(sizeof(struct IFile) + strlen(logname));
	p->prv = ictxt.fin;
	ictxt.fin = p;

	line = 0;
	strcpy(fnam, logname);

	if(logname != fname) {		/* fname == "-" --> stdin */
		f = stdin;
		return 0;
	}

	if((f = fopen(fname, "rb")) != NULL)
		return 0;

	popFile();
	return 1;
}

static void pushFile(char *fname)
{	if(probeFile(fname) != 0)
		unf_error UNF_EOpenFile, "Cannot open file: \"%s\"", fname);
}

#define endTable()	d->data[tablepos] = e, numSizes = 0

#define ifOptWrite() if(isReadonly(&opts[i].flags, makeReadonly))
static int isReadonly(int *flags, int mkrdo)
{	int isrdo;

	isrdo = *flags & READONLY;
	if(mkrdo)
		*flags |= READONLY;

	return !isrdo;
}

struct UNF_DATA *unf_load(char *path, UWORD cntry, UWORD cp, UNF_Error_Fct *fct)
{	char *buf, *p, *q;
	int state;
	struct UNF_DATA *d;
	unsigned offset, disp;		/* virtual offset within buffer */
	struct UNF_DATA_ENTRY *e;
	int numSizes;			/* how many "size" are maintained */
	unsigned *sizes;		/* where the sizes are */
	int tablepos;

		/* Initialize input context */
	errid = UNF_ENone;
	ictxt.error_fct = fct? fct: (UNF_Error_Fct*)unf_dummy_error;
	ictxt.fin = NULL;

		/* Reset READONLY flag within the static table */
	tablepos = itemsof(opts);
	while(tablepos--)
		opts[tablepos].flags &= ~READONLY;

	if(!path)
		path = "";

	if(probeFile(path) != 0) {			/* failed with given name */
		/* try to append a name */
		buf = unf_alloc(strlen(path) + 1 + 2*(sizeof(cntry) * 8 + 1) + 1 + 4);
		if(*path) {
			p = unf_stpcpy(buf, path);
			if(p[-1] != '/' && p[-1] != '\\' && p[-1] != ':')
				*p++ = '\\';
		} else p = buf;
		if(cntry > 999 && cp > 999) {
			unf_error UNF_ENumExceeded, "Both country ID and codepage exceed 999");
			/** free(buf); is performed below **/
			goto errRet;
		}
		sprintf(p, "%03u-%03u.unf", cntry, cp);
		pushFile(buf);
		free(buf);
	}

	buf = unf_alloc(UNF_BUFSIZ);
	d = unf_alloc(sizeof(struct UNF_DATA));
	d->country = d->codepage = (UWORD)UNF_DEFAULT;
	d->numData = 0;
	irx = 10;
	state = GLOBAL;
	numSizes = 0;
	sizes = NULL;
	do {
		while(fgets(p = buf, UNF_BUFSIZ, f)) {
			iferr goto errRet;
			++line;
			if(!strchr(p, '\n')) {
				unf_error UNF_ELineTooLong, "Line too long");
				goto errRet;
			}
			if((q = strchr(p, '#')) == NULL)
				q = strchr(p, '\0');
			while(--q >= p && isspace(*q));
			q[1] = '\0';
			
			switch(*(p = skipws(p))) {
			case '$':		/* option */
				{	int i, ch;
					char opt[MAX_OPT_SIZE + 1];
					int makeReadonly;

					if((makeReadonly = p[1] == '$') != 0)
						++p;
					for(i = 0; i < MAX_OPT_SIZE && (isalpha(*++p) || *p == '_')
					 ; ++i)
						opt[i] = tolower(*p);

					if(i == 0 || i == MAX_OPT_SIZE || (*p && !isspace(*p))) {
						unf_error UNF_EInvalidOpt, "Invalid option or name too long");
						goto errRet;
					}
					opt[i] = '\0';
					i = 0;
					while(strcmp(opt, opts[i].name) != 0)
						if(++i >= itemsof(opts)) {
							unf_error UNF_EUnknownOpt, "Unknown option: \"%s\"");
							goto errRet;
						}
					if((opts[i].flags & state) == 0)	/* not in this section */
						break;

					p = skipws(p);

					switch(opts[i].id) {
					case OPT_Radix:
						ifOptWrite() {
							/* check for special radix's */
							if(!p[1]) switch(toupper(*p)) {
							case 'B': irx = UNF_RDX_COMBINED; goto nxt;
							case 'C': irx = UNF_RDX_CSTYLE; goto nxt;
							}
							irx = 10;	/* radix of $radix is always 10 */
							getWord(irx);
							if(irx < 2 || irx > 36) {
								unf_error UNF_ERadix, "The radix must be in range 2..36");
								goto errRet;
							}
						}
					nxt:
						break;
					case OPT_Country:
						ifOptWrite() {
							getWord(d->country);
						}
						break;
					case OPT_Codepage:
						ifOptWrite() {
							getWord(d->codepage);
						}
						break;
					case OPT_ChainTo:
						popFile();
					case OPT_Include:
						pushFile(p);
						break;
					case OPT_IsOffset:
						getWord(disp);
						if((numSizes? *(UWORD*)&(e->data[sizes[0]])
						 : offset) != disp) {
							unf_error UNF_EOffMismatch, "Offset mismatch");
							goto errRet;
						}
						break;
					case OPT_MakeSize:
						addValue(2, 1, 0);
						sizes = unf_realloc(sizes
						 , ++numSizes * sizeof(unsigned));
						sizes[numSizes - 1] = offset - 2;
						break;
					case OPT_StopSize:
						numSizes = 0;
						free(sizes);
						sizes = NULL;
						break;
					case OPT_Goto:
						getWord(offset);
#if 1
                        expandBuf(offset);
#endif
						break;
					case OPT_GoBack:
						getWord(disp);
						if(offset >= disp)
							offset -= disp;
						else offset = 0;
						break;
					case OPT_GoForward:
						getWord(disp);
						offset += disp;
#if 1
                        expandBuf(offset);
#endif
						break;
					case OPT_GoEven:
						if(offset & 1)
							++offset;
						break;
					case OPT_GoOdd:
						if((offset & 1) == 0)
							++offset;
						break;
					}
				}
				break;
			case '=':		/* data line */
				if(state != LOCAL) {
					unf_error UNF_EMisplacedData, "Data lines are not allowed here");
					goto errRet;
				}
				++p;
				while(p && *p) if(isspace(*p))
					++p;
				else {
					int size, count;
					unsigned long value;

					switch(*p++) {
					default:
						--p;
					case 'b':	case 'B':
						size = 1;		/* one byte */
						break;
					case 'w':	case 'W':
						size = 2;		/* two byte */
						break;
					case 'd': case 'D':
						size = 4;		/* four byte */
						break;
					}

					if(*(p = inum(p, &value)) == '\0' || isspace(*p))
						count = 1;
					else if(*p == '*') {
						count = (int)value;
						if(*(p = inum(p, &value)) != '\0' && !isspace(*p)) {
							unf_error UNF_EDataline, "Syntax error in data line %lu", line);
							goto errRet;
						}
					} else {
						unf_error UNF_EDataline, "Syntax error in data line %lu", line);
						goto errRet;
					}

					addValue(size, count, value);
				}
				break;
			case '[':		/* new table */
				if(state == GLOBAL) {
					if(d->country == (UWORD)UNF_DEFAULT
					 || d->codepage == (UWORD)UNF_DEFAULT) {
					 	unf_error UNF_EMissingBasedata, "Missing country ID or codepage");
					 	goto errRet;
					}
					state = LOCAL;
				} else {
					endTable();
				}
				p = skipws(p + 1);
				if(memicmp("table", p, 5) != 0
				 || !isspace(*(p += 5))) {
					unf_error UNF_EInvalidTableName, "Invalid table name");
					goto errRet;
				}
				{	unsigned long num;

					if(*skipws(inum(p, &num)) != ']' || num > 255) {
						unf_error UNF_EInvalidTableName, "Invalid table name");
						goto errRet;
					}
					for(tablepos = 0; tablepos < d->numData; ++tablepos)
						if(d->data[tablepos]->fct == (int)num) {
							free(d->data[tablepos]);
							if(tablepos < --d->numData)
								memmove(&d->data[tablepos]
								 , &d->data[tablepos + 1]
								 , (d->numData - tablepos)
								   * sizeof(struct UNF_DATA_ENTRY*));
							break;
						}
					e = unf_alloc(sizeof(struct UNF_DATA_ENTRY));
					e->fct = (UBYTE)num;
					offset = 0;
				}
				e->size = 0;
					/* sizeof(UNF_DATA) contains 1 DATA_ENTRY */
				d = unf_realloc(d, sizeof(struct UNF_DATA)
				 + d->numData * sizeof(struct UNF_DATA_ENTRY*));
				tablepos = d->numData++;
				break;
			case '\0':		/* empty --> comment */
				break;
			default:
				unf_error UNF_ESyntax, "Syntax error");
				goto errRet;
			}
		}
	} while(popFile());

errRet:
	if(state != GLOBAL)
		endTable();

	while(popFile());
	free(sizes);
	free(buf);

	ifnerr return d;		/* On success */

	/* On error remove the structure and all allocated memory */
	unf_freeData(d);
	return NULL;
}
